#include <bits/stdc++.h>
#include <omp.h>
#include <mpi.h>

#define SUBDOMAIN_SIZE 100
#define NUM_ITERS 10

#define TAG_SOUTH 0
#define TAG_NORTH 1

using namespace std;

int main(int argc, char* argv[])
{

    int rank, size;
    int errcnt, toterr;
    double sub_local[SUBDOMAIN_SIZE+2][SUBDOMAIN_SIZE];

    MPI_Init( &argc, &argv );

    MPI_Comm_rank(MPI_COMM_WORLD, &rank);
    MPI_Comm_size(MPI_COMM_WORLD, &size);

    // 1. Init subdomain with the rank ID
    for (int i = 0; i <= SUBDOMAIN_SIZE; i++){
        for (int j = 0; j <= SUBDOMAIN_SIZE; j++){
            sub_local[i][j] = rank;
        }
    }
    // 2. Init halo to -1
    for (int j = 0; j < SUBDOMAIN_SIZE; j++){
        sub_local[0][j] = -1;
        sub_local[SUBDOMAIN_SIZE+1][j] = -1;
    }

    for(int iterations=0; iterations < NUM_ITERS; ++iterations){
        if(rank >= 1){
            // Send south
            MPI_Send(&sub_local[1][0], SUBDOMAIN_SIZE, MPI_DOUBLE, rank-1, TAG_SOUTH, MPI_COMM_WORLD);
            // Receive north
            MPI_Recv(&sub_local[0][0], SUBDOMAIN_SIZE, MPI_DOUBLE, rank-1, TAG_NORTH, MPI_COMM_WORLD, MPI_STATUS_IGNORE);
        }
        if(rank < size - 1){
            // Send north
            MPI_Send(&sub_local[SUBDOMAIN_SIZE][0], SUBDOMAIN_SIZE, MPI_DOUBLE, rank+1, TAG_NORTH, MPI_COMM_WORLD);
            // Receive south
            MPI_Recv(&sub_local[SUBDOMAIN_SIZE+1][0], SUBDOMAIN_SIZE, MPI_DOUBLE, rank+1, TAG_SOUTH, MPI_COMM_WORLD, MPI_STATUS_IGNORE);
        }
        errcnt = 0;
        
        for (int i = 1; i <= SUBDOMAIN_SIZE; i++) 
            for (int j = 0; j < SUBDOMAIN_SIZE; j++) 
                if (sub_local[i][j] != rank) errcnt++;

        for (int j = 0; j < SUBDOMAIN_SIZE; j++){
            if (sub_local[0][j] != rank - 1) errcnt++;
            if (rank < size - 1 && sub_local[SUBDOMAIN_SIZE+1][j] != rank + 1) errcnt++;
        }
        
        MPI_Reduce(&errcnt, &toterr, 1, MPI_INT, MPI_SUM, 0, MPI_COMM_WORLD);
        
        if(rank == 0){
            cout <<"End of iteration: " << iterations << " - Total errors: " << toterr << endl;
        }   
    }

    return 0;
}
