#include <bits/stdc++.h>
#include <omp.h>
#include <mpi.h>

#define SUBDOMAIN_SIZE 200
#define NUM_ITERS 10

#define TAG_SOUTH 0
#define TAG_NORTH 1

using namespace std;

int main(int argc, char* argv[])
{

    int rank, size;
    double sub_local[SUBDOMAIN_SIZE+2][SUBDOMAIN_SIZE];
    double sub_local_aux[SUBDOMAIN_SIZE+2][SUBDOMAIN_SIZE];

    MPI_Init( &argc, &argv );

    MPI_Comm_rank(MPI_COMM_WORLD, &rank);
    MPI_Comm_size(MPI_COMM_WORLD, &size);

    // 1. Init subdomain with the rank ID
    for (int i = 0; i <= SUBDOMAIN_SIZE; i++){
        for (int j = 0; j <= SUBDOMAIN_SIZE; j++){
            sub_local[i][j] = static_cast<double>(rand()) / static_cast<double>(RAND_MAX);
        }
    }
    // 2. Init halo to -1
    for (int j = 0; j < SUBDOMAIN_SIZE; j++){
        sub_local[0][j] = -1;
        sub_local[SUBDOMAIN_SIZE+1][j] = -1;
    }

    for(int iterations=0; iterations < NUM_ITERS; ++iterations){
        // HALO EXCHANGE:
        MPI_Request request[4];
        MPI_Status status[4];

        if(rank >= 1){
            // Send south
            MPI_Isend(&sub_local[1][0], SUBDOMAIN_SIZE, MPI_DOUBLE, rank-1, TAG_SOUTH, MPI_COMM_WORLD, &request[0]);
            // Receive north
            MPI_Irecv(&sub_local[0][0], SUBDOMAIN_SIZE, MPI_DOUBLE, rank-1, TAG_NORTH, MPI_COMM_WORLD, &request[1]);
        }
        if(rank < size - 1){
            // Send north
            MPI_Isend(&sub_local[SUBDOMAIN_SIZE][0], SUBDOMAIN_SIZE, MPI_DOUBLE, rank+1, TAG_NORTH, MPI_COMM_WORLD, &request[2]);
            // Receive south
            MPI_Irecv(&sub_local[SUBDOMAIN_SIZE+1][0], SUBDOMAIN_SIZE, MPI_DOUBLE, rank+1, TAG_SOUTH, MPI_COMM_WORLD, &request[3]);
        }
        
        // COMPUTATION: BODY
        #pragma omp parallel for
        for (int i = 2; i < SUBDOMAIN_SIZE; i++){
            for (int j = 0; j < SUBDOMAIN_SIZE; j++){
                sub_local_aux[i][j] = (sub_local[i][j] + sub_local[i-1][j] + sub_local[i+1][j] + sub_local[i][j-1] + sub_local[i][j+1]) / 5.0;
            }
        }
        
        // COMPUTATION: NORTH AND SOUTH -> WAIT Irecv to be finished!
        if(rank >= 1){
            MPI_Wait(&request[1], &status[1]);
        }
        if(rank < size - 1){
            MPI_Wait(&request[3], &status[3]);
        }
        #pragma omp parallel for
        for (int j = 0; j < SUBDOMAIN_SIZE; j++){
            sub_local_aux[1][j] = (sub_local[1][j] + sub_local[0][j] + sub_local[2][j] + sub_local[1][j-1] + sub_local[1][j+1]) / 5.0;
            sub_local_aux[SUBDOMAIN_SIZE][j] = (sub_local[SUBDOMAIN_SIZE][j] + sub_local[SUBDOMAIN_SIZE-1][j] + sub_local[SUBDOMAIN_SIZE+1][j] + sub_local[SUBDOMAIN_SIZE][j-1] + sub_local[SUBDOMAIN_SIZE][j+1]) / 5.0;
        }

        // COPY AUX TO LOCAL -> WAIT Isend to be finished!
        if(rank >= 1){
            MPI_Wait(&request[0], &status[0]);
        }
        if(rank < size - 1){
            MPI_Wait(&request[2], &status[2]);
        }
        for (int i = 1; i <= SUBDOMAIN_SIZE; i++){
            for (int j = 0; j < SUBDOMAIN_SIZE; j++){
                sub_local[i][j] = sub_local_aux[i][j];
            }
        }
        

        if(rank == 0){
            cout <<"End of iteration: " << iterations << endl;
        }   
    }

    return 0;
}
